#define INCL_WIN
#define INCL_DOS
#define INCL_GPIBITMAPS
#include <os2.h>

#include <string.h>

#include "dive.h"
#include "mglserver_internal.h"
#include "controls.h"

#define TIMEOUT 5 // seconds before giving up on the MGL server daemon

#define MGLSC_WAKEUP_SEM_NAME (PUCHAR)"\\SEM32\\MGLS_PM_WAKEUP"
#define MGLSC_LISTENER_NAME   (PUCHAR)"\\QUEUES\\MGLS_CLIENT_LISTENER"

#define FOURCC_BGR4 0x34524742ul
#define FOURCC_RGB4 0x34424752ul
#define FOURCC_RGB3 0x33424752ul
#define FOURCC_R565 0x35363552ul
#define FOURCC_LUT8 0x3854554cul
#define FOURCC_SCRN 0

struct MGLSC_state          mglclient;
struct MGLSC_event_handlers mglevents;

char MGLup = 0, mouseGrab = 0, mouseRel = 1;
int mouseAbsX, mouseAbsY, mouseCurX, mouseCurY;
int PMCenterX, PMCenterY;
char mouseMoved = 0, mouseUngrabbedByFocus = 0;

char MGLSC_PMkeyToMGLKeycode( char keycode )
{
  return PMkeys[ keycode ].MGL_key;
}

char MGLSC_MGLkeyToPMKeycode( char keycode )
{
  int i;

  for ( i=0; i<DEFINED_KEYS; ++i )
  {
    if ( PMkeys[ i ].MGL_key == keycode )
    {
      return i;
    }
  }
  return 0; // Illegal scan code in both MGL and PM
}

const char * MGLSC_getKeyName( char PMkeycode )
{
  return PMkeys[ PMkeycode ].description;
}

struct MGLSC_state *MGLSC_clientState( void )
{
  return &mglclient;
}

void MGLSC_flushUserInput( void )
{
  if ( !mglclient.isFullScreen )
  {
    if ( mglevents.ProcessInput && mouseGrab && mouseMoved )
    {
      POINTL ppos;
      int tmpCurX;
      int tmpCurY;

      WinQueryPointerPos( HWND_DESKTOP, &ppos );
      mouseCurX = ppos.x;
      mouseCurY = ppos.y;

      tmpCurX = mouseCurX - PMCenterX;
      tmpCurY = PMCenterY - mouseCurY;
      // "Up" increases row in PM

      if ( !tmpCurX && !tmpCurY )
      {
        mouseMoved = 0;
        return;
      }
      // No real movement happened.

      // Make sure PM mouse input is acted on too

      if ( mouseRel )
      {
        if ( tmpCurX >  mglclient.width/2 )  tmpCurX =   mglclient.width/2;
        if ( tmpCurX < -mglclient.width/2 )  tmpCurX = -(mglclient.width/2);
        if ( tmpCurY >  mglclient.height/2 ) tmpCurY =   mglclient.height/2;
        if ( tmpCurY < -mglclient.height/2 ) tmpCurY = -(mglclient.height/2);

        mglevents.ProcessInput( MGLSC_MOUSE_MOVE, tmpCurX, tmpCurY );
      } else {
        mouseAbsX += tmpCurX;
        mouseAbsY += tmpCurY;

        if ( mouseAbsX >= mglclient.width ) mouseAbsX = mglclient.width - 1;
        if ( mouseAbsX < 0 ) mouseAbsX = 0;
        if ( mouseAbsY >= mglclient.height ) mouseAbsY = mglclient.height - 1;
        if ( mouseAbsY < 0 ) mouseAbsY = 0;

        mglevents.ProcessInput( MGLSC_MOUSE_MOVE, mouseAbsX, mouseAbsY );
      }
      mouseCurX = PMCenterX;
      mouseCurY = PMCenterY;
      mouseMoved = 0;
      WinPostMsg( mglclient.clientwin, WM_MOUSE_RECENTER, 0, 0 );
    }

    return;
  }

  DosWriteQueue( mglclient.command_queue, MGLS_EVENT_FLUSH, 0, NULL, 0 );
  if ( hEJoy && mglevents.ProcessInput )
  {
    int i;

    // update joystick input too

    for ( i=0; i<num_joysticks; ++i )
    {
      ejoy_ext_get_info( i );
      if ( !i )
      {
        mglevents.ProcessInput( MGLSC_JOYSTICK_MOVE, ejoy_info.xpos, ejoy_info.ypos );
        mglevents.ProcessInput( MGLSC_JOYSTICK_BUTTON, ejoy_info.buttons, 0 );
      } else {
        mglevents.ProcessInput( MGLSC_JOYSTICK2_MOVE, ejoy_info.xpos, ejoy_info.ypos );
        mglevents.ProcessInput( MGLSC_JOYSTICK2_BUTTON, ejoy_info.buttons, 0 );
      }
    }
  }
}

void MGLS_synchronous_command( int command, int packetsize, void *packet )
{
  ULONG junk;

  DosResetEventSem( mglclient.client_wakeup, &junk );
  if ( !mglclient.isFullScreen ) return;
  DosWriteQueue( mglclient.command_queue, command, packetsize, packet, 0 );
  DosWaitEventSem( mglclient.client_wakeup, 2000 );
}

static void MGLSC_MGL_reinit( void );

void ClientListenerThread( void *unused )
{
  REQUESTDATA request;
  ULONG information, rc, junk;
  BYTE priority;
  static int lastAbsX = -1, lastAbsY = -1;

  DosSetPriority( PRTYS_THREAD, PRTYC_TIMECRITICAL, 0, 0 );
  // Make sure keyboard is responsive

  rc = DosReadQueue( mglclient.listener_queue, &request, &junk,
    (void **) &information, 0, 0, &priority, 0 );

  while ( !rc ) // While queue handle is valid
  {
    switch ( request.ulData )
    {
      case MGLC_VIDEO_SWITCH_NOTIFICATION:
        mglclient.isFullScreen = information;
        mglclient.palette_dirty = 1;
        // Make sure HW palette is refreshed when we change
        // video modes

        #ifdef MGLSC_DEBUG_LOG
          fprintf( mglclient.dbfp, "Video mode switch notification received.  Mode = %d\n", information );
          fflush( mglclient.dbfp );
        #endif

        if ( !mglclient.isFullScreen )
        {
          DosPostEventSem( mglclient.client_wakeup );
        } else {
          if ( mglclient.vidmode_reinit )
          {
            mglclient.disallow_blit = 1;
            MGLSC_MGL_reinit();
            mglclient.disallow_blit = 0;
          }
          // Video mode changed and this is the first time the MGL session
          // was reactivated, so make the change take effect now.

          if ( mglclient.autoBlitVidBufferOnRefresh )
          {
            WinPostMsg( mglclient.clientwin, WM_BLITFRAME, 0, 0 );
          }
        }
      break;
      case MGLC_KEYDOWN_NOTIFICATION:
        if ( mglevents.ProcessInput )
          mglevents.ProcessInput( MGLSC_KEYBOARD_MAKE, information, 0 );
      break;
      case MGLC_KEYUP_NOTIFICATION:
        if ( mglevents.ProcessInput )
          mglevents.ProcessInput( MGLSC_KEYBOARD_BREAK, information, 0 );
      break;
      case MGLC_MOUSEMOVE_NOTIFICATION:
        if ( mglevents.ProcessInput )
        {
          if ( mouseRel )
          {
            if ( lastAbsX < 0 )
            {
              lastAbsX = mglclient.width / 2;
              lastAbsY = mglclient.height / 2;
            }
            // MGLServer returns absolute coordinates, so relativize them here
            mglevents.ProcessInput( MGLSC_MOUSE_MOVE,
             (information & 0xffff) - lastAbsX,
             ((information >> 16) & 0xffff) - lastAbsY );
            lastAbsX = information & 0xffff;
            lastAbsY = (information >> 16) & 0xffff;
          } else {
            mglevents.ProcessInput( MGLSC_MOUSE_MOVE, information & 0xffff,
             (information >> 16) & 0xffff );
          }
        }
      break;
      case MGLC_MOUSEMAKE_NOTIFICATION:
        if ( mglevents.ProcessInput )
          mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_PRESS, information, 0 );
      break;
      case MGLC_MOUSEBREAK_NOTIFICATION:
        if ( mglevents.ProcessInput )
          mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_RELEASE, information, 0 );
      break;
    }

    rc = DosReadQueue( mglclient.listener_queue, &request, &junk, 
      (void **) &information, 0, 0, &priority, 0 );
  }

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Client listener queue has shut down.  DosReadQueue rc=%d.\n", rc );
    fflush( mglclient.dbfp );
  #endif

  DosPostEventSem( mglclient.client_wakeup );
  // Tell the main thread that we're done.

  _endthread();
}

int MGLSC_MGL_init( void )
{
  ULONG rc, i;
  STARTDATA sd = {0};
  RESULTCODES res;
  MGL_SERVER_INIT_PACKET *ip;
  MGL_SERVER_INIT_VIDMODE_PACKET *iv;
  MGL_SERVER_INIT_BUFFER_PACKET *ib;
  STATUSDATA statdat;

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Create wakeup sem\n" );
    fflush( mglclient.dbfp );
  #endif

  rc = DosCreateEventSem( MGLSC_WAKEUP_SEM_NAME, 
    &(mglclient.client_wakeup), DC_SEM_SHARED, FALSE );

  if ( rc )
  {
    return 1;
  }

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Check if server is already running\n" );
    fflush( mglclient.dbfp );
  #endif

  rc = DosOpenQueue( &(mglclient.server_pid), &(mglclient.command_queue),
    MGLS_COMMAND_QUEUE_NAME );

  if ( !rc )
  {
		DosCloseQueue( mglclient.command_queue );
    DosCloseEventSem( mglclient.client_wakeup );
    return 2;
  }
  // Daemon is already running and is probably owned by someone else.
  // Don't start up.

  if ( rc != 343 )
  {
    return 3;
  }
  // DosOpenQueue failed for some other wacky reason.

  sd.Length = 32;
  sd.Related = 1;
  sd.FgBg = SSF_FGBG_FORE;
  sd.TraceOpt = SSF_TRACEOPT_NONE;
  sd.PgmTitle = NULL;
  sd.PgmName = (PUCHAR)"mglserver.exe";
  sd.PgmInputs = NULL;
  sd.TermQ = NULL;
  sd.Environment = 0;
  sd.InheritOpt = SSF_INHERTOPT_PARENT;
  sd.SessionType = SSF_TYPE_FULLSCREEN;

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Start up new server\n" );
    fflush( mglclient.dbfp );
  #endif

  rc = DosStartSession( &sd, &(mglclient.MGL_session), &mglclient.server_pid );

  if ( rc ) { 
    DosCloseQueue( mglclient.command_queue );
    DosCloseEventSem( mglclient.client_wakeup );
    return 4;
  }

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Set session properties\n" );
    fflush( mglclient.dbfp );
  #endif

  statdat.Length = sizeof( USHORT ) * 3;
  statdat.SelectInd = SET_SESSION_NON_SELECTABLE;
  statdat.BondInd = 0;
  DosSetSession( mglclient.MGL_session, &statdat );
  // Take the ugly "MGLSERVER.EXE" out of the "Switch To" list

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Wait for command queue to exist\n" );
    fflush( mglclient.dbfp );
  #endif

  for (i=0; i<TIMEOUT; ++i)
  {
    DosSleep(1000);
    rc = DosOpenQueue( &(mglclient.server_pid), &(mglclient.command_queue),
      MGLS_COMMAND_QUEUE_NAME );
    if ( !rc ) break;
  }

  if ( rc )
  {
    DosCloseEventSem( mglclient.client_wakeup );
    return 6;
  }

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Create listener queue\n" );
    fflush( mglclient.dbfp );
  #endif

  rc = DosCreateQueue( &(mglclient.listener_queue), QUE_FIFO, MGLSC_LISTENER_NAME );
  if ( rc )
  {
    DosCloseQueue( mglclient.command_queue );
    DosCloseEventSem( mglclient.client_wakeup );
    DosFreeMem( mglclient.shared_packet );
    return 7;
  }

  ip = (MGL_SERVER_INIT_PACKET *) mglclient.shared_packet;
  strcpy( ip->client_semaphore_name, (char *)(MGLSC_WAKEUP_SEM_NAME) );
  strcpy( ip->input_queue_name, (char *)(MGLSC_LISTENER_NAME) );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Launch listener thread\n" );
    fflush( mglclient.dbfp );
  #endif

  mglclient.listener_thread = _beginthread( 
    ClientListenerThread, NULL, 16384, NULL );

  // Fire off the listener thread to get feedback from the
  // daemon.  Needed for mode switch notification and input
  // device notifications.

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Client listener thread (%d) started.\n", 
      mglclient.listener_thread );
    fprintf( mglclient.dbfp, "Go full screen\n" );
    fflush( mglclient.dbfp );
  #endif

  rc = DosSelectSession( mglclient.MGL_session );
  if ( rc )
  {
    DosKillThread( mglclient.listener_thread );
    DosCloseQueue( mglclient.command_queue );
    DosCloseQueue( mglclient.listener_queue );
    DosCloseEventSem( mglclient.client_wakeup );
    DosFreeMem( mglclient.shared_packet );
    return 8;
  }

  mglclient.isFullScreen = 1;

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Init server\n" );
    fflush( mglclient.dbfp );
  #endif

  MGLS_synchronous_command( MGLS_INIT, sizeof(MGL_SERVER_INIT_PACKET), ip );

  iv = (MGL_SERVER_INIT_VIDMODE_PACKET *) mglclient.shared_packet;
  iv->flags = 0;

  for ( i=0; mglclient.allowedmodes[i].width; ++i )
  {
    if ( mglclient.allowedmodes[i].width < mglclient.width ) continue;
    if ( mglclient.allowedmodes[i].height < mglclient.height ) continue;
    if ( mglclient.allowedmodes[i].depth < mglclient.depth ) continue;

    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "Attempting FS mode %d x %d %dbpp.\n", 
        mglclient.allowedmodes[i].width, mglclient.allowedmodes[i].height,
        mglclient.allowedmodes[i].depth );
      fflush( mglclient.dbfp );
    #endif

    iv->width = mglclient.allowedmodes[i].width;
    iv->height = mglclient.allowedmodes[i].height;
    iv->depth = mglclient.allowedmodes[i].depth;
    iv->flags = (mglclient.custommodes ? MGLS_VMIFLAG_USE_CUSTOM : 0);

    MGLS_synchronous_command( MGLS_INIT_VIDMODE, 
      sizeof(MGL_SERVER_INIT_VIDMODE_PACKET), iv );

    if ( iv->flags & MGLS_VMOFLAG_SUCCESS ) break;
  }

  if ( !(iv->flags & MGLS_VMOFLAG_SUCCESS) )
  {
    MGLS_synchronous_command( MGLS_SHUTDOWN,
      sizeof(MGL_SERVER_SHUTDOWN_PACKET), mglclient.shared_packet );

    DosSelectSession( 0 );

    DosKillThread( mglclient.listener_thread );
    DosFreeMem( mglclient.shared_packet );
    DosCloseQueue( mglclient.command_queue );
    DosCloseQueue( mglclient.listener_queue );

    DosCloseEventSem( mglclient.client_wakeup );
    return 9;
    // Could not use any acceptable modes
  }

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Success using FS mode %d x %d %dbpp.\n", 
      mglclient.allowedmodes[i].width, mglclient.allowedmodes[i].height,
      mglclient.allowedmodes[i].depth );
    fflush( mglclient.dbfp );
  #endif

  mglclient.fswidth  = mglclient.allowedmodes[i].width;
  mglclient.fsheight = mglclient.allowedmodes[i].height;

  ib = (MGL_SERVER_INIT_BUFFER_PACKET *) mglclient.shared_packet;
  ib->buffer = mglclient.vidbuffer;
  ib->width = mglclient.width;
  ib->height = mglclient.height;
  ib->depth = mglclient.depth;

  MGLS_synchronous_command( MGLS_INIT_BUFFER, 
    sizeof(MGL_SERVER_INIT_BUFFER_PACKET), ib );
  // MGL now has a memory device context associated with this buffer.
  // We're ready to blit!

  mglclient.autoBlitVidBufferOnRefresh = 1;

  return 0;
}

void MGLS_shutdown( void )
{
  ULONG junk;

  DosSelectSession( mglclient.MGL_session );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Telling MGLServer to free buffer.\n" );
    fflush( mglclient.dbfp );
  #endif

  MGLS_synchronous_command( MGLS_FREE_BUFFER,
    sizeof(MGL_SERVER_FREE_BUFFER_PACKET), mglclient.shared_packet );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Telling MGLServer to shut down video mode.\n" );
    fflush( mglclient.dbfp );
  #endif

  MGLS_synchronous_command( MGLS_SHUTDOWN_VIDMODE,
    sizeof(MGL_SERVER_SHUTDOWN_VIDMODE_PACKET), mglclient.shared_packet );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Telling MGLServer itself to shut down.\n" );
    fflush( mglclient.dbfp );
  #endif

  MGLS_synchronous_command( MGLS_SHUTDOWN,
    sizeof(MGL_SERVER_SHUTDOWN_PACKET), mglclient.shared_packet );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Free up video buffer and shared packet storage.\n" );
    fflush( mglclient.dbfp );
  #endif

  DosFreeMem( mglclient.shared_packet );
  DosFreeMem( mglclient.vidbuffer );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Close command and listener queues.\n" );
    fflush( mglclient.dbfp );
  #endif

  DosCloseQueue( mglclient.command_queue );
  DosCloseQueue( mglclient.listener_queue );

  DosCloseEventSem( mglclient.client_wakeup );

  DosSleep( 1000 );
  // Take a breather

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "MGLServer shutdown succeeded.\n" );
    fflush( mglclient.dbfp );
  #endif
}

MRESULT EXPENTRY DualModePM(HWND win, ULONG message, MPARAM mp1, MPARAM mp2)
{
  static ULONG err;
  static HPS hps;
  static HRGN hrgn;
  static RGNRECT rgnCtl;
  static RECTL rcls[50];
  static SWP swp;
  static POINTL pointl;

  switch ( message )
  {
    case WM_CREATE:
    {
      MGLup = 0;
      mglclient.autoBlitVidBufferOnRefresh = 1;

      WinQueryWindowRect( HWND_DESKTOP, &(rcls[0]) );
      PMCenterX = (rcls[0].xRight - rcls[0].xLeft) / 2;
      PMCenterY = (rcls[0].yTop - rcls[0].yBottom) / 2;

      err = DiveOpen( &(mglclient.diveinst), FALSE, NULL );
      if ( err ) {
        return MRFROMSHORT(-1);
      }

      err = DiveAllocImageBuffer( mglclient.diveinst, 
        &(mglclient.divebufnum), mglclient.depthfourcc,
        mglclient.width, mglclient.height, 
        mglclient.width * (mglclient.depth / 8), 
        (PBYTE *) mglclient.vidbuffer );

      if ( err ) {
        DiveClose( mglclient.diveinst );
        mglclient.diveinst = 0;
        return MRFROMSHORT(-1);
      }
      break;
    }
    case WM_BLITFRAME:
    {
      if ( mglclient.disallow_blit ) return 0;
      if ( mglclient.isFullScreen )
      {
        MGL_SERVER_BLIT_BUFFER_PACKET *bb = 
          (MGL_SERVER_BLIT_BUFFER_PACKET *)
          mglclient.shared_packet;

        if ( mglclient.palette_dirty )
        {
          MGL_SERVER_COLORS_SET_PACKET *cs = 
            (MGL_SERVER_COLORS_SET_PACKET *)
            mglclient.shared_packet;

          // Palette should already be loaded into cs->colors[]

          DosWriteQueue( mglclient.command_queue, MGLS_SET_COLORS,
            sizeof(MGL_SERVER_COLORS_SET_PACKET), cs, 0 );
          // Don't sync this call, just sync the frame blit below.
          // (no need to waste twice the time)
        }

        bb->left = 0;
        bb->right = mglclient.width;
        bb->top = 0;
        bb->bottom = mglclient.height;
        if ( mglclient.width  < mglclient.fswidth ||
             mglclient.height < mglclient.fsheight )
        {
          if ( mglclient.stretchblit )
          {
            MGL_SERVER_STRETCHBLIT_BUFFER_PACKET *sb = 
              (MGL_SERVER_STRETCHBLIT_BUFFER_PACKET *)
              mglclient.shared_packet;

            sb->destleft = 0;
            sb->desttop = 0;
            sb->destright = mglclient.fswidth;
            sb->destbottom = mglclient.fsheight;

            MGLS_synchronous_command( MGLS_STRETCHBLIT_BUFFER,
              sizeof(MGL_SERVER_BLIT_BUFFER_PACKET), sb );
            // Stretch image to screen size
          } else {
            bb->destx = (mglclient.fswidth  - mglclient.width)  / 2;
            bb->desty = (mglclient.fsheight - mglclient.height) / 2;
            // Center on the screen

            MGLS_synchronous_command( MGLS_BLIT_BUFFER,
              sizeof(MGL_SERVER_BLIT_BUFFER_PACKET), bb );
          }
        } else {
          bb->destx = 0;
          bb->desty = 0;

          MGLS_synchronous_command( MGLS_BLIT_BUFFER,
            sizeof(MGL_SERVER_BLIT_BUFFER_PACKET), bb );
        }
      } else {
        if ( mglclient.palette_dirty )
        {
          MGL_SERVER_COLORS_SET_PACKET *cs = 
            (MGL_SERVER_COLORS_SET_PACKET *)
            mglclient.shared_packet;

          DiveSetSourcePalette( mglclient.diveinst, 0,
            256, (PBYTE)cs->colors );
        }
        DiveBlitImage( mglclient.diveinst, mglclient.divebufnum, 
          DIVE_BUFFER_SCREEN );
      }
      mglclient.palette_dirty = 0;
      if ( mglclient.framesync_sem )
        DosPostEventSem( mglclient.framesync_sem );
      break;
    }
    case WM_CHAR:
    {
      if ( (SHORT1FROMMP( mp1 ) & KC_SCANCODE) )
      {
        if ( (SHORT1FROMMP( mp1 ) & KC_KEYUP) == 0 )
        {
          if ( mglevents.ProcessInput )
            mglevents.ProcessInput( MGLSC_KEYBOARD_MAKE,
             PMkeys[ CHAR4FROMMP(mp1) ].MGL_key, 0 );
        } else {
          if ( mglevents.ProcessInput )
            mglevents.ProcessInput( MGLSC_KEYBOARD_BREAK,
             PMkeys[ CHAR4FROMMP(mp1) ].MGL_key, 0 );
        }
        // Some scancodes are different between MGL and PM.
        // Xlate as needed
      }
      break;
    }
    case WM_CLOSE:
      if ( hEJoy ) ejoy_close();

      if ( mglevents.WindowClosed )
        mglevents.WindowClosed();

      break;
    case WM_PAINT:
      if ( !mglclient.disallow_blit )
      {
        DiveBlitImage( mglclient.diveinst, mglclient.divebufnum,
         DIVE_BUFFER_SCREEN );
      }
      break;
    case WM_VRNENABLE:
      if ( !mglclient.diveinst ) break;
      hps = WinGetPS(win);
      if ( hps == 0 ) { err=1; break; }
      hrgn=GpiCreateRegion(hps, 0L, NULL);
      if (hrgn) {
        err = WinQueryVisibleRegion(win, hrgn);
        rgnCtl.ircStart=0;
        rgnCtl.crc=50;
        rgnCtl.ulDirection=RECTDIR_LFRT_TOPBOT;
        err = GpiQueryRegionRects(hps, hrgn, NULL, &rgnCtl, rcls);
        if ( err ) { 
          err = 0; 
          WinQueryWindowPos(win, &swp);
          pointl.x=swp.x;
          pointl.y=swp.y;
          WinMapWindowPoints(WinQueryWindow( win, QW_PARENT ), HWND_DESKTOP,
            (POINTL*)&pointl, 1);
          mglclient.BlSet.ulStructLen = sizeof( SETUP_BLITTER );
          mglclient.BlSet.fInvert = 0;
          mglclient.BlSet.fccSrcColorFormat = mglclient.depthfourcc;
          mglclient.BlSet.ulSrcWidth = mglclient.width;
          mglclient.BlSet.ulSrcHeight = mglclient.height;
          mglclient.BlSet.ulSrcPosX = 0;
          mglclient.BlSet.ulSrcPosY = 0;
          mglclient.BlSet.ulDitherType = 1;
          mglclient.BlSet.fccDstColorFormat = FOURCC_SCRN;
          mglclient.BlSet.ulDstWidth=swp.cx;
          mglclient.BlSet.ulDstHeight=swp.cy;
          mglclient.BlSet.lDstPosX = 0;
          mglclient.BlSet.lDstPosY = 0;
          mglclient.BlSet.lScreenPosX=pointl.x;
          mglclient.BlSet.lScreenPosY=pointl.y;
          mglclient.BlSet.ulNumDstRects=rgnCtl.crcReturned;
          mglclient.BlSet.pVisDstRects=rcls;

          if ( rgnCtl.crcReturned == 0 )
          { 
            DiveSetupBlitter( mglclient.diveinst, NULL );
            GpiDestroyRegion(hps, hrgn);
            WinReleasePS( hps );
            return MRFROMSHORT(-1);
          }
          err = DiveSetupBlitter( mglclient.diveinst,
            &(mglclient.BlSet) );
					if ( err )
          {
            #ifdef MGLSC_DEBUG_LOG
              fprintf( mglclient.dbfp, "VRNENABLE failed %x.\n", err );
              fflush( mglclient.dbfp );
            #endif
            return MRFROMSHORT(-1);
          }

          GpiDestroyRegion(hps, hrgn);
        }
      }
      WinReleasePS( hps );
      WinPostMsg( win, WM_PAINT, NULL, NULL );
    break;
    case WM_VRNDISABLE:
      DiveSetupBlitter( mglclient.diveinst, NULL );
    break;
    case WM_TOGGLEFS:
    {
      ULONG crap;

      if ( mglclient.isFullScreen )
      {
        #ifdef MGLSC_DEBUG_LOG
          fprintf( mglclient.dbfp, "Switching to PM.  (sesID=0)\n" );
          fflush( mglclient.dbfp );
        #endif

        DosSetPriority( PRTYS_THREAD, PRTYC_REGULAR, 0, 1 );
        // Make sure the foreground boost is undone when we switch back from
        // full screen mode.

        DosSelectSession( 0 );
      } else {
        int rc;

        if ( !MGLup )
        {
          #ifdef MGLSC_DEBUG_LOG
            fprintf( mglclient.dbfp, "Initializing MGL server.\n" );
            fflush( mglclient.dbfp );
          #endif
          rc = MGLSC_MGL_init();
          if ( rc )
          {
            #ifdef MGLSC_DEBUG_LOG
              fprintf( mglclient.dbfp, "MGL server startup error #%d.\n", rc );
              fflush( mglclient.dbfp );
            #endif
            break;
          }
          MGLup = 1;
        } else {
          int rc;
          #ifdef MGLSC_DEBUG_LOG
            fprintf( mglclient.dbfp, "Switching to FS.  (sesID=%ld)\n", mglclient.MGL_session );
            fflush( mglclient.dbfp );
          #endif
          rc = DosSelectSession( mglclient.MGL_session );
          if ( rc )
          {
            #ifdef MGLSC_DEBUG_LOG
              fprintf( mglclient.dbfp, "DosSelectSession returned %d!\n", rc );
            #endif
          }
          DosSetPriority( PRTYS_THREAD, PRTYC_FOREGROUNDSERVER, 0, 1 );
          // Make sure the process *generating* the graphics still gets its
          // foreground priority boost, like it would if it were the PM session
          // with the window focus.
        }
      }
    }
    break;
    case WM_COMMAND:
      if ( mglevents.PMMenuAction )
        mglevents.PMMenuAction( SHORT1FROMMP( mp1 ) );
    break;
    case WM_CLIPCAPTURE:
    {
      BITMAPINFOHEADER2 bmpi = {0};
      HBITMAP bmp;
      HPS hps;
      DEVOPENSTRUC dop = { 0, (PUCHAR)"DISPLAY", NULL, 0, 0, 0, 0, 0, 0 };
      SIZEL size = {0, 0};
      HDC hdc;
      unsigned char *bitmap24bit;
      int i,j;

      hdc = DevOpenDC( WinQueryAnchorBlock( mglclient.clientwin ),
       OD_MEMORY, (PUCHAR)"*", 5, (PDEVOPENDATA)&dop, NULLHANDLE );
      hps = GpiCreatePS( WinQueryAnchorBlock( mglclient.clientwin ),
       hdc, &size, PU_PELS | GPIA_ASSOC );

      bmpi.cbFix = sizeof(BITMAPINFOHEADER2);
      bmpi.cx = mglclient.width;
      bmpi.cy = mglclient.height;
      bmpi.cPlanes = 1;
      bmpi.cBitCount = 24;

      bitmap24bit = (unsigned char *)
       malloc( mglclient.width * mglclient.height * 3 );

      if ( mglclient.depth == 16 )
      {
        unsigned short *vidbuffer = mglclient.vidbuffer;
        for ( i=0; i<mglclient.height; ++i )
        {
          for ( j=0; j<mglclient.width; ++j )
          {
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 0 ] =
              (vidbuffer[ (i*mglclient.width) + j ] & 0x1f) << 3;
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 1 ] =
              ((vidbuffer[ (i*mglclient.width) + j ] >> 5) & 0x3f) << 2;
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 2 ] =
              ((vidbuffer[ (i*mglclient.width) + j ] >> 11) & 0x1f) << 3;
          }
        }
      } else if ( mglclient.depth == 24 )
      {
        memcpy( bitmap24bit, mglclient.vidbuffer,
         mglclient.width * mglclient.height * 3 );
      } else if ( mglclient.depth == 32 )
      {
        unsigned long *vidbuffer = mglclient.vidbuffer;
        for ( i=0; i<mglclient.height; ++i )
        {
          for ( j=0; j<mglclient.width; ++j )
          {
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 0 ] = vidbuffer[ (i*mglclient.width) + j ] & 0xff;
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 1 ] = (vidbuffer[ (i*mglclient.width) + j ] >> 8) & 0xff;
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 2 ] = (vidbuffer[ (i*mglclient.width) + j ] >> 16) & 0xff;
          }
        }
      } else {
        MGL_SERVER_COLORS_SET_PACKET *cs = mglclient.shared_packet;
        unsigned char *vidbuffer = mglclient.vidbuffer;

        for ( i=0; i<mglclient.height; ++i )
        {
          for ( j=0; j<mglclient.width; ++j )
          {
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 2 ] =
              cs->colors[ vidbuffer[ (i*mglclient.width) + j ] ].red;
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 1 ] =
              cs->colors[ vidbuffer[ (i*mglclient.width) + j ] ].green;
            bitmap24bit[ (((mglclient.height-1)-i) * mglclient.width*3)
             + (j*3) + 0 ] =
              cs->colors[ vidbuffer[ (i*mglclient.width) + j ] ].blue;
          }
        }
      }

      bmp = GpiCreateBitmap( hps, &bmpi, CBM_INIT, bitmap24bit,
       (BITMAPINFO2 *) &bmpi );
      free( bitmap24bit );

      if ( !bmp )
      {
        #ifdef MGLSC_DEBUG_LOG
          fprintf( mglclient.dbfp, "Error capturing bitmap!  GpiCreateBitmap error 0x%lx.\n",
           WinGetLastError( WinQueryAnchorBlock( mglclient.clientwin ) ) );
        #endif
        WinAlarm( HWND_DESKTOP, WA_ERROR );
        GpiDestroyPS(hps);
        DevCloseDC(hdc);
        return;
      }

      WinOpenClipbrd( WinQueryAnchorBlock( mglclient.clientwin ) );
      WinSetClipbrdOwner( WinQueryAnchorBlock( mglclient.clientwin ),
       mglclient.clientwin );
      WinEmptyClipbrd( WinQueryAnchorBlock( mglclient.clientwin ) );
      WinSetClipbrdData( WinQueryAnchorBlock( mglclient.clientwin ),
       bmp, CF_BITMAP, CFI_HANDLE );
      WinAlarm( HWND_DESKTOP, WA_WARNING );

      WinSetClipbrdOwner( WinQueryAnchorBlock( mglclient.clientwin ),
       NULLHANDLE );
      // Clipboard takes ownership of the handle and is responsible for
      // deleting the resources associated with it.

      WinCloseClipbrd( WinQueryAnchorBlock( mglclient.clientwin ) );
      GpiDestroyPS( hps );
      DevCloseDC( hdc );
    }
    break;
    case WM_MOUSEMOVE:
      if ( mouseGrab )
      {
        mouseMoved = 1;
        WinSetPointer( HWND_DESKTOP, 0 );
        return;
      }
    break;
    case WM_MOUSE_RECENTER:
      WinSetPointerPos( HWND_DESKTOP, PMCenterX, PMCenterY );
    break;
    case WM_BUTTON1DOWN:
      if ( mglevents.ProcessInput )
        mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_PRESS, 1, 0 );
    break;
    case WM_BUTTON1UP:
      if ( mglevents.ProcessInput )
        mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_RELEASE, 1, 0 );
    break;
    case WM_BUTTON2DOWN:
      if ( mglevents.ProcessInput )
        mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_PRESS, 2, 0 );
    break;
    case WM_BUTTON2UP:
      if ( mglevents.ProcessInput )
        mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_RELEASE, 2, 0 );
    break;
    case WM_BUTTON3DOWN:
      if ( mglevents.ProcessInput )
        mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_PRESS, 3, 0 );
    break;
    case WM_BUTTON3UP:
      if ( mglevents.ProcessInput )
        mglevents.ProcessInput( MGLSC_MOUSE_BUTTON_RELEASE, 3, 0 );
    break;
    case WM_FOCUSCHANGE:
      if ( mouseGrab && !SHORT1FROMMP(mp2) )
      {
        // Mouse is grabbed but we're losing the window focus.
        // Ungrab the mouse so the user can control their system.

        WinSetCapture( HWND_DESKTOP, 0 );
        mouseGrab = 0;
        mouseUngrabbedByFocus = 1;
      } else if ( mouseUngrabbedByFocus && SHORT1FROMMP(mp2) )
      {
        // Mouse grabbing was disabled by a focus change, but the focus
        // has changed back, so reenable grabbing.

        WinSetCapture( HWND_DESKTOP, mglclient.clientwin );
        WinSetPointer( HWND_DESKTOP, 0 );
        mouseGrab = 1;
        mouseUngrabbedByFocus = 0;
      }
    break;
    case WM_SETWINSIZE:
    {
      RECTL framer, desktopr;
      ULONG swpopts = SWP_SIZE;
      USHORT scale = SHORT1FROMMP(mp1);

      if ( SHORT1FROMMP( mp2 ) ) swpopts |= SWP_MOVE;

      framer.xLeft = 0;
      framer.xRight = mglclient.width * scale;
      framer.yBottom = 0;
      framer.yTop = mglclient.height * scale;
      WinCalcFrameRect( mglclient.framewin, &framer, FALSE );
      // Calculate needed frame rect from give client rect

      WinQueryWindowRect( HWND_DESKTOP, &desktopr );

      WinPostMsg( mglclient.clientwin, WM_VRNDISABLE, NULL, NULL );
      WinSetWindowPos( mglclient.framewin, HWND_TOP,
        (((desktopr.xRight-desktopr.xLeft)/2)+desktopr.xLeft) -
        ((framer.xRight-framer.xLeft)/2),
        (((desktopr.yTop-desktopr.yBottom)/2)+desktopr.yBottom) -
        ((framer.yTop-framer.yBottom)/2),
        framer.xRight - framer.xLeft,
        framer.yTop - framer.yBottom,
        swpopts );
      WinPostMsg( mglclient.clientwin, WM_VRNENABLE, NULL, NULL );
    }
    break;
  }

  if ( mglevents.AdditionalWindowProc )
  {
    return mglevents.AdditionalWindowProc( win, message, mp1, mp2 );
  } else {
    return WinDefWindowProc( win, message, mp1, mp2 );
  }
}


void PMThread( void )
{
  HAB ab;
  HMQ messq;
  QMSG qmsg;
  ULONG frameflgs= FCF_SIZEBORDER;
  RECTL framer, desktopr;

  if ( mglclient.windowdecor & MGLSC_WINDOW_DECOR_MENU )
   frameflgs |= FCF_MENU;
  if ( mglclient.windowdecor & MGLSC_WINDOW_DECOR_ICON )
   frameflgs |= FCF_ICON;
  if ( mglclient.windowdecor & MGLSC_WINDOW_DECOR_TASKLIST )
   frameflgs |= FCF_TASKLIST;
  if ( mglclient.windowdecor & MGLSC_WINDOW_DECOR_MINMAX )
   frameflgs |= FCF_MINMAX;
  if ( mglclient.windowdecor & MGLSC_WINDOW_DECOR_SYSMENU )
   frameflgs |= FCF_SYSMENU;
  if ( mglclient.windowdecor & MGLSC_WINDOW_DECOR_TITLEBAR )
   frameflgs |= FCF_TITLEBAR;

  ab = WinInitialize( 0 );

  messq = WinCreateMsgQueue( ab, 0 );

  WinRegisterClass( ab, (PUCHAR)mglclient.winclassname, DualModePM, 
    CS_SIZEREDRAW | CS_MOVENOTIFY, 0 );
        
  mglclient.framewin = WinCreateStdWindow( HWND_DESKTOP, WS_ANIMATE,
    &frameflgs, (PUCHAR)mglclient.winclassname,
    (PUCHAR)mglclient.wintitle, 0, 0, 1, &(mglclient.clientwin) );

  framer.xLeft = 0;
  framer.xRight = mglclient.width;
  framer.yBottom = 0;
  framer.yTop = mglclient.height;
  WinCalcFrameRect( mglclient.framewin, &framer, FALSE );
  // Calculate needed frame rect from give client rect

  WinQueryWindowRect( HWND_DESKTOP, &desktopr );

  WinPostMsg( mglclient.clientwin, WM_VRNDISABLE, NULL, NULL );
  WinSetWindowPos( mglclient.framewin, HWND_TOP,
    (((desktopr.xRight-desktopr.xLeft)/2)+desktopr.xLeft) -
    ((framer.xRight-framer.xLeft)/2),
    (((desktopr.yTop-desktopr.yBottom)/2)+desktopr.yBottom) -
    ((framer.yTop-framer.yBottom)/2),
    framer.xRight - framer.xLeft,
    framer.yTop - framer.yBottom,
    SWP_SIZE | SWP_MOVE );
  // Set 1:1 scaling and center

  if ( mglevents.WindowInitializer )
    mglevents.WindowInitializer();
  // Let the application customize the window as desired

  WinShowWindow( mglclient.framewin, TRUE );
  WinSetFocus( HWND_DESKTOP, mglclient.clientwin );
  WinPostMsg( mglclient.clientwin, WM_VRNENABLE, NULL, NULL );

  WinSetVisibleRegionNotify( mglclient.clientwin, TRUE );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Entering message queue loop in PM thread.\n" );
    fflush( mglclient.dbfp );
  #endif

  while (WinGetMsg (ab, &qmsg, NULLHANDLE, 0, 0))
  {
    WinDispatchMsg (ab, &qmsg);
  }

  mglclient.autoBlitVidBufferOnRefresh = 0;

  if ( mglclient.diveinst )
  {
    if ( mglclient.divebufnum )
    {
      DiveFreeImageBuffer( mglclient.diveinst, mglclient.divebufnum );
      mglclient.divebufnum = 0;
    }
    DiveClose( mglclient.diveinst );
    mglclient.diveinst = 0;
  }

  if ( MGLup ) 
  {
    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "Switching to MGL session for shutdown.\n" );
      fflush( mglclient.dbfp );
    #endif

    DosSelectSession( mglclient.MGL_session );
    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "Shutting down MGL Server.\n" );
      fflush( mglclient.dbfp );
    #endif
    MGLS_shutdown();
    MGLup = 0;
  }

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Destroying window.\n" );
    fflush( mglclient.dbfp );
  #endif

  WinDestroyWindow( mglclient.framewin );
  mglclient.framewin = mglclient.clientwin = 0;

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Destroying message queue.\n" );
    fflush( mglclient.dbfp );
  #endif

  WinDestroyMsgQueue( messq );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Terminating anchor block.\n" );
    fflush( mglclient.dbfp );
  #endif

  WinTerminate( ab );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Finished.  Closing debug file.\n" );
    fclose( mglclient.dbfp );
    mglclient.dbfp = NULL;
  #endif

}

void MGLSC_init( int width, int height, int depth,
 struct MGLSC_event_handlers *eventhandler, char *winclassname,
 char *wintitle, struct vidmode_descrip *available_modes,
 char windowDecorations, char allowCustomModes,
 char useJoystickIfAvailable )
{
  int i;

  memset( &mglclient, 0, sizeof( mglclient ) );

  mglclient.windowdecor = windowDecorations;

  #ifdef MGLSC_DEBUG_LOG
    mglclient.dbfp = fopen( "mglsc_debug.log", "wa" );
  #endif

  mglclient.allowedmodes = available_modes;
  mglclient.custommodes = allowCustomModes;

  memcpy( &mglevents, eventhandler, sizeof( struct MGLSC_event_handlers ) );

  mglclient.winclassname = winclassname;
  mglclient.wintitle = wintitle;

  switch ( depth )
  {
    case 8:   mglclient.depthfourcc = FOURCC_LUT8; break;
    case 16:  mglclient.depthfourcc = FOURCC_R565; break;
    case 24:  mglclient.depthfourcc = FOURCC_RGB3; break;
    case 32:  mglclient.depthfourcc = FOURCC_RGB4; break;
    default:  return;  // Unsupported depth;
  }

  mglclient.width = width;
  mglclient.height = height;
  mglclient.depth = depth;

  mglclient.isFullScreen = 0;

  DosAllocSharedMem( &(mglclient.vidbuffer), NULL, 
    width * height * (depth/8), OBJ_GETTABLE | PAG_READ | PAG_WRITE | PAG_COMMIT );

  DosAllocSharedMem( &(mglclient.shared_packet), NULL, 
    MGLS_BIGGEST_PACKET_SIZE, OBJ_GETTABLE | PAG_READ | PAG_WRITE | PAG_COMMIT );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "Starting PM Thread.\n" );
    fflush( mglclient.dbfp );
  #endif

  mglclient.pm_thread =
    _beginthread( (void(*)(void*))PMThread, NULL, 32768, NULL );

  #ifdef MGLSC_DEBUG_LOG
    fprintf( mglclient.dbfp, "PM Thread ID = %d.\n", mglclient.pm_thread );
    fflush( mglclient.dbfp );
  #endif

  if ( useJoystickIfAvailable )
  {
    if ( !ejoy_open() )
    {
      #ifdef MGLSC_DEBUG_LOG
        fprintf( mglclient.dbfp, "Joystick driver opened.\n" );
        fflush( mglclient.dbfp );
      #endif

      if ( !ejoy_check_version() )
      {
        #ifdef MGLSC_DEBUG_LOG
          fprintf( mglclient.dbfp, "Unsupported version of joystick driver.\n" );
          fflush( mglclient.dbfp );
        #endif

        ejoy_close();
        hEJoy = 0;
      }
    }
  }

  mouseAbsX = width / 2;
  mouseAbsY = height / 2;

  while ( !mglclient.divebufnum ) DosSleep( 500 );
  // Lazy man's semaphore.

  // DIVE is fully initialized now.  It's safe to return.
}

void MGLSC_reinit( int width, int height, int depth, char allowCustomModes,
 char useJoystickIfAvailable )
{
  RECTL framer;
  ULONG err;

  mglclient.custommodes = allowCustomModes;

  if ( !useJoystickIfAvailable && hEJoy )
  {
    ejoy_close();
    hEJoy = 0;
  } else if ( useJoystickIfAvailable && !hEJoy )
  {
    if ( !ejoy_open() )
    {
      #ifdef MGLSC_DEBUG_LOG
        fprintf( mglclient.dbfp, "Joystick driver opened.\n" );
        fflush( mglclient.dbfp );
      #endif

      if ( !ejoy_check_version() )
      {
        #ifdef MGLSC_DEBUG_LOG
          fprintf( mglclient.dbfp, "Unsupported version of joystick driver.\n" );
          fflush( mglclient.dbfp );
        #endif

        ejoy_close();
        hEJoy = 0;
      }
    }
  }

  if ( width != mglclient.width || height != mglclient.height ||
       depth != mglclient.depth )
  {
    // Screen mode change

    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "Request to change video mode to %d x %d %dbpp.\n", 
        width, height, depth );
      fflush( mglclient.dbfp );
    #endif

    DiveSetupBlitter( mglclient.diveinst, NULL );
    // Stop DIVE blitting while we reallocate buffer sizes

    mglclient.disallow_blit = 1;
    // Stop MGL & DIVE blits

    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "DIVE and MGL blitters have been disabled.\n" );
      fflush( mglclient.dbfp );
    #endif

    DiveFreeImageBuffer( mglclient.diveinst, mglclient.divebufnum );
    mglclient.divebufnum = 0;
    DosFreeMem( mglclient.vidbuffer );

    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "Old video buffer has been freed.\n" );
      fflush( mglclient.dbfp );
    #endif

    switch ( depth )
    {
      case 8:
        mglclient.depthfourcc = FOURCC_LUT8;
        break;
      case 16:
        mglclient.depthfourcc = FOURCC_R565;
        break;
      case 24:
        mglclient.depthfourcc = FOURCC_RGB3;
        break;
      case 32:
        mglclient.depthfourcc = FOURCC_RGB4;
        break;
      default:
        return;
      // Unsupported color depth
    }

    DosAllocSharedMem( &(mglclient.vidbuffer), NULL,
     width * height * (depth/8), OBJ_GETTABLE | PAG_READ | PAG_WRITE |
     PAG_COMMIT );

    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "New video buffer allocated.\n" );
      fflush( mglclient.dbfp );
    #endif

    mglclient.width = width;
    mglclient.height = height;
    mglclient.depth = depth;

    err = DiveAllocImageBuffer( mglclient.diveinst, 
      &(mglclient.divebufnum), mglclient.depthfourcc,
      width, height, width * (depth / 8), 
      (PBYTE *) mglclient.vidbuffer );

    if ( err )
    {
      #ifdef MGLSC_DEBUG_LOG
        fprintf( mglclient.dbfp, "ERROR!: DiveAllocImageBuffer returned %lx.\n", err );
        fflush( mglclient.dbfp );
      #endif

      DiveClose( mglclient.diveinst );
      mglclient.diveinst = 0;
      return;
    }

    WinPostMsg( mglclient.clientwin, WM_SETWINSIZE, MPFROMSHORT(1), 0 );
    // Resize the window if needed

    // This should take care of the PM/DIVE aspect, now deal with making
    // the changes to the MGL session if it is up.

    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "DIVE settings were changed.\n" );
      fflush( mglclient.dbfp );
    #endif

    if ( MGLup )
    {
      if ( !mglclient.isFullScreen )
      {
        mglclient.vidmode_reinit = 1;
        // Have to wait until the MGL session is reactivated to change full
        // screen video mode.
        mglclient.disallow_blit = 0;
        // Allow the window to blit in the meantime.
      } else {
        MGLSC_MGL_reinit();
        mglclient.disallow_blit = 0;
        // re-allow blitting
      }
    } else {
      mglclient.disallow_blit = 0;
      // Since MGL isn't up yet, we should be all set by this point to
      // begin blitting again.
    }
    WinPostMsg( mglclient.clientwin, WM_VRNENABLE, 0, 0 );
  }
}

static void MGLSC_MGL_reinit( void )
{
  int i;
  MGL_SERVER_INIT_VIDMODE_PACKET *iv;
  MGL_SERVER_INIT_BUFFER_PACKET *ib;

  MGLS_synchronous_command( MGLS_FREE_BUFFER,
    sizeof(MGL_SERVER_FREE_BUFFER_PACKET), mglclient.shared_packet );
  MGLS_synchronous_command( MGLS_SHUTDOWN_VIDMODE,
    sizeof(MGL_SERVER_SHUTDOWN_VIDMODE_PACKET), mglclient.shared_packet );
  // Gets us back to text mode and frees up old resources

  iv = (MGL_SERVER_INIT_VIDMODE_PACKET *) mglclient.shared_packet;
  iv->flags = 0;

  for ( i=0; mglclient.allowedmodes[i].width; ++i )
  {
    if ( mglclient.allowedmodes[i].width < mglclient.width ) continue;
    if ( mglclient.allowedmodes[i].height < mglclient.height ) continue;
    if ( mglclient.allowedmodes[i].depth < mglclient.depth ) continue;

    #ifdef MGLSC_DEBUG_LOG
      fprintf( mglclient.dbfp, "Attempting FS mode %d x %d %dbpp.\n", 
        mglclient.allowedmodes[i].width, mglclient.allowedmodes[i].height,
        mglclient.allowedmodes[i].depth );
      fflush( mglclient.dbfp );
    #endif

    iv->width = mglclient.allowedmodes[i].width;
    iv->height = mglclient.allowedmodes[i].height;
    iv->depth = mglclient.allowedmodes[i].depth;
    iv->flags = (mglclient.custommodes ? MGLS_VMIFLAG_USE_CUSTOM : 0);

    MGLS_synchronous_command( MGLS_INIT_VIDMODE, 
      sizeof(MGL_SERVER_INIT_VIDMODE_PACKET), iv );

    if ( iv->flags & MGLS_VMOFLAG_SUCCESS ) break;
  }

  if ( !(iv->flags & MGLS_VMOFLAG_SUCCESS) )
  {
    MGLS_synchronous_command( MGLS_SHUTDOWN,
      sizeof(MGL_SERVER_SHUTDOWN_PACKET), mglclient.shared_packet );

    DosSelectSession( 0 );

    DosKillThread( mglclient.listener_thread );
    DosFreeMem( mglclient.shared_packet );
    DosCloseQueue( mglclient.command_queue );
    DosCloseQueue( mglclient.listener_queue );

    DosCloseEventSem( mglclient.client_wakeup );
    return;
    // Could not use any acceptable modes
  }

  ib = (MGL_SERVER_INIT_BUFFER_PACKET *) mglclient.shared_packet;
  ib->buffer = mglclient.vidbuffer;
  ib->width = mglclient.width;
  ib->height = mglclient.height;
  ib->depth = mglclient.depth;

  MGLS_synchronous_command( MGLS_INIT_BUFFER, 
    sizeof(MGL_SERVER_INIT_BUFFER_PACKET), ib );

  mglclient.vidmode_reinit = 0;
  mglclient.disallow_blit = 0;
  // Video mode is all set so reset these flags
}

char MGLSC_isJoystickPresent( void )
{
  return enhanced_joystick_interface;
}

char MGLSC_getNumJoysticks( void )
{
  ejoy_ext_get_caps( 0 );
  return ejoy_caps.nro_of_joysticks;
}

char MGLSC_getNumJoyButtons( LONG whichStick )
{
  return num_joysticks;
}

void MGLSC_mouseGrab( char yesNo )
{
  mouseCurX = 0;
  mouseCurY = 0;

  mouseUngrabbedByFocus = 0;

  if ( (mouseGrab && !yesNo) || (!mouseGrab && yesNo) )
  {
    mouseGrab = yesNo;

    if ( !mouseGrab )
    {
      WinSetCapture( HWND_DESKTOP, 0 );
    } else {
      WinSetCapture( HWND_DESKTOP, mglclient.clientwin );
      WinSetPointer( HWND_DESKTOP, 0 );
    }
  }
}

void MGLSC_setMouseModeRelative( char isRelative )
{
  mouseRel = isRelative;
  if ( !isRelative )
  {
    mouseAbsX = mglclient.width / 2;
    mouseAbsY = mglclient.height / 2;

    mglevents.ProcessInput( MGLSC_MOUSE_MOVE, mouseAbsX, mouseAbsY );    
    // Make sure client knows that the mouse absolute position has been
    // re-centered the next time it asks to flush the event queue.
  }
}

